class Segment:
    def __init__(self,id,feature_id,chr,start,stop,strand):
        self.id=id
        self.feature_set=set([(feature_id,0)]) # set of tuples (feat_id,index_in_info_list), for all features on this segment. if the feature is on the + strand, it will be named '+feature_id', else '-feature_id'

        self.feat_info_list=[[chr,int(start),int(stop),strand]] # position on the original genome for the features on this segment

    # add a feature to an existing segment
    def add_feature(self,line,new_feat_id,new_strand):
        if not self.find_feat(new_feat_id)[0]:
            feat_info=[line[0],int(line[1])+1,int(line[2]),new_strand] # +1 in the start to convert the bed 0-based coordinate to a 1-based system

            if feat_info in self.feat_info_list:
                feat_tuple=new_feat_id,self.feat_info_list.index(feat_info)
                self.feature_set.add(feat_tuple)
            else:
                feat_tuple=(new_feat_id,len(self.feat_info_list))
                self.feat_info_list.append(feat_info)
                self.feature_set.add(feat_tuple)

    def find_feat(self,feat_id):
        for feat_tuple in self.feature_set:
            if feat_tuple[0]==feat_id:
                return (True,feat_tuple[1])
        return (False,0)
    
    def get_start(self,feat_id):
        feat_present=self.find_feat(feat_id)
        if feat_present[0]:
            index_feat=feat_present[1]
            return self.feat_info_list[index_feat][1]

    def get_stop(self,feat_id):
        feat_present=self.find_feat(feat_id)
        if feat_present[0]:
            index_feat=feat_present[1]
            return self.feat_info_list[index_feat][2]

    def get_chr(self,feat_id):
        feat_present=self.find_feat(feat_id)
        if feat_present[0]:
            index_feat=feat_present[1]
            return self.feat_info_list[index_feat][0]

    def get_strand(self,feat_id):
        feat_present=self.find_feat(feat_id)
        if feat_present[0]:
            index_feat=feat_present[1]
            return self.feat_info_list[index_feat][3]

    def get_size(self,feat_id):
        feat_present=self.find_feat(feat_id)
        if feat_present[0]:
            return self.get_stop(feat_id)-self.get_start(feat_id)+1

    def __str__(self):
        return f"id={self.id}, features={self.feature_set}, features info={self.feat_info_list}"


class Feature:
    def __init__(self,id,type,chr,start,stop,annot,childs,seg_list,strand,complete,discontinuous):
        self.id=id
        self.type=type
        
        # position on the original genome
        self.chr=chr
        self.start=start
        self.stop=stop
        self.strand=strand
        self.size=stop-start+1

        self.pos_start=0 # position on the first segment, setup later
        self.pos_stop=0 # position on the last segment, setup later
        
        self.annot=annot
        self.dict_annot=self.build_dict_annot()
        if 'Parent' in self.dict_annot:
            self.parent=self.dict_annot["Parent"].replace(".","_").replace(":","_")
        else:
            self.parent=""
        self.childs=childs # list of child features (exons, cds, etc)
        self.segments_list_source=seg_list # list of oriented segments on which the feature is (>1/<1, depending on the path of the gene in the graph)
        self.segments_list_target=[] # list of lists, as there can be several occurences of a feature on a given target genome

        self.note="" # from the annotation file
        self.sequence=""

        self.complete=complete # boolean, is the feature initialised with all the info or not

        self.discontinuous=discontinuous # boolean, is the feature discontinuous in the gff (split in several lines, like a fragmented cds for instance)
        self.first='' # boolean, is it the first part of the feature encountered
        self.first_part='' # if first=False, string, the id of the first part encountered
        self.other_parts_list=[] # if first=True, list ot the id of the other parts encountered

    # returns a list of feature's child, and their childs' childs, etc. 
    def get_child_list(self,Features):
        list_childs=[]
        for child in self.childs:
            list_childs.append(child) # add the child to the list
            list_childs+=child.get_child_list(Features) # add the child's childs to the list
        return list_childs
    
    # create a note for the child features that do not have annotation.
    def set_note(self):
        # the note contains information on the function of the feature and is used for statistics on hypothetical/putatives features.
        # in the gff, the notes are only on the "gene" features. it's easier to have it for the childs than to check the parent's note (or the parent's parent). 
        note=""
        parent=self.find_parent()
        if parent!=None: # in case find_parent() returns nothing; should never happen
            if "Note" in parent.dict_annot:
                note=parent.dict_annot["Note"]
        self.note=note

    # find the gene parent of a feature
    def find_parent(self): # add a check if there is no parent !
        if self.parent=='': # no parent, usually a gene
            return self
        else:
            current=self.parent
            parent_found=False
            while parent_found==False:
                if current.parent=='': # current doesnt have a parent
                    return current
                else: # if we didn't find the parent, we go up to the current feature's parent until we find it
                    current=current.parent

    # find on what part of the source feature the segment is; sometimes the features are fragmented.
    def find_part_segment_source(self,segment):
        if segment in self.segments_list_source:
            return self
        else:
            list_parts=self.other_parts_list
            for part in list_parts:
                if segment in part.segments_list_source:
                    return part
        
    # add a segment to an existing feature
    def add_seg(self,new_seg_oriented):
        # if new_seg_oriented in Features[feat_id].segments_list_source:
        #     print("seg already in feat, duplicate segment ! feature:",feat_id,"segment:",new_seg_oriented)
        #     print("list for now :",Features[feat_id].segments_list_source)
        self.segments_list_source.append(new_seg_oriented)

    def build_dict_annot(self):
        dict_annot={}
        if self.annot=='':
            return dict_annot
        list_fields=self.annot.split(";")
        for field in list_fields:
            [key,value]=field.split("=")
            dict_annot[key]=value
        return dict_annot

    def __str__(self):
        if self.parent=="":
            return f"id={self.id}, type={self.type}, segments={self.segments_list_source}, position on the original genome={self.chr}:{self.start}-{self.stop}, childs={len(self.childs)}, annotation={self.annot}"
        else:
            return f"id={self.id}, type={self.type}, segments={self.segments_list_source}, position on the original genome={self.chr}:{self.start}-{self.stop}, parent={self.parent.id}, childs={len(self.childs)}, annotation={self.annot}"